import django.core.validators
from django.db import migrations, models
import yubival.models
import yubival.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='APIKey',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(default=yubival.models.generate_api_key, max_length=28, validators=[yubival.validators.LengthValidator(28)])),
            ],
        ),
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('public_id', models.CharField(default=yubival.models.generate_public_id, max_length=12, unique=True, validators=[yubival.validators.LengthValidator(12), yubival.validators.validate_modhex])),
                ('private_id', models.CharField(default=yubival.models.generate_private_id, max_length=12, unique=True, validators=[yubival.validators.LengthValidator(12), yubival.validators.validate_hex])),
                ('key', models.CharField(default=yubival.models.generate_otp_key, max_length=32, validators=[yubival.validators.LengthValidator(32), yubival.validators.validate_hex])),
                ('session_counter', models.IntegerField(default=0, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(32767)])),
                ('usage_counter', models.IntegerField(default=0, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(255)])),
            ],
        ),
    ]



from django.db import migrations, models
import django.utils.timezone


def set_apikey_labels(apps, schema_editor):
    APIKey = apps.get_model('yubival', 'APIKey')
    db_alias = schema_editor.connection.alias
    for api_key in APIKey.objects.using(db_alias).all():
        api_key.label = 'Label %d' % api_key.id
        api_key.save()


def set_device_labels(apps, schema_editor):
    Device = apps.get_model('yubival', 'Device')
    db_alias = schema_editor.connection.alias
    for device in Device.objects.using(db_alias).all():
        device.label = 'Label %d' % device.id
        device.save()


class Migration(migrations.Migration):

    dependencies = [
        ('yubival', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='apikey',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),

        migrations.AddField(
            model_name='apikey',
            name='label',
            field=models.CharField(max_length=64, null=True),
            preserve_default=False,
        ),
        migrations.RunPython(set_apikey_labels, migrations.RunPython.noop),
        migrations.AlterField(
            model_name='apikey',
            name='label',
            field=models.CharField(max_length=64, unique=True, null=False),
            preserve_default=False,
        ),

        migrations.AddField(
            model_name='device',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),

        migrations.AddField(
            model_name='device',
            name='label',
            field=models.CharField(max_length=64, null=True),
            preserve_default=False,
        ),
        migrations.RunPython(set_device_labels, migrations.RunPython.noop),
        migrations.AlterField(
            model_name='device',
            name='label',
            field=models.CharField(max_length=64, unique=True, null=False),
            preserve_default=False,
        ),
    ]



import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('yubival', '0002_add_label_and_date_created_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='device',
            name='session_counter',
            field=models.IntegerField(default=0, editable=False, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(32767)]),
        ),
        migrations.AlterField(
            model_name='device',
            name='usage_counter',
            field=models.IntegerField(default=0, editable=False, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(255)]),
        ),
    ]



import string

from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible


def argparse_type(field, func_name=None):
    """Creates an argparse type corresponding to a Django model field

    Returns a function that can be used as an argparse parameter type that performs input validation and conversion
    that correspond to a given Django model field.

    Args:
        field: a Django database model field class such as IntegerField.
        func_name: name of the type function that may be shown in argparse parsing error messages.

    Returns:
        type_func: function that converts its input value to a value with a data type consistent with the field provided
            above. If there are validation errors, a `ValueError` is raised.
    """

    def type_func(value):
        value = field.get_prep_value(value)

        for v in field.validators:
            try:
                v(value)
            except ValidationError:
                raise ValueError

        return value

    if func_name is not None:
        type_func.__name__ = func_name

    return type_func


@deconstructible
class LengthValidator:
    def __init__(self, length):
        self.length = length

    def __call__(self, value):
        if len(value) != self.length:
            raise ValidationError("Must be of length %d." % self.length)


def validate_hex(value):
    if not all(c in string.hexdigits for c in value):
        raise ValidationError("Must be composed of hexadecimal characters.")


def validate_modhex(value):
    if not all(c in 'cbdefghijklnrtuv' for c in value):
        raise ValidationError("Must be composed of modhex characters.")



import base64
import datetime
import hashlib
import hmac
from binascii import hexlify
from collections import OrderedDict
from enum import Enum

from django.db import transaction
from django.http import HttpResponse
from django.utils.http import urlencode
from django.views import View
from yubiotp.otp import decode_otp

from yubival.models import APIKey, Device


class ValidationStatus(Enum):
    OK = 'OK'
    BAD_OTP = 'BAD_OTP'
    REPLAYED_OTP = 'REPLAYED_OTP'
    BAD_SIGNATURE = 'BAD_SIGNATURE'
    MISSING_PARAMETER = 'MISSING_PARAMETER'
    NO_SUCH_CLIENT = 'NO_SUCH_CLIENT'
    OPERATION_NOT_ALLOWED = 'OPERATION_NOT_ALLOWED'
    BACKEND_ERROR = 'BACKEND_ERROR'
    NOT_ENOUGH_ANSWERS = 'NOT_ENOUGH_ANSWERS'
    REPLAYED_REQUEST = 'REPLAYED_REQUEST'


def ordered_parameters(params):
    ordered_params = OrderedDict()
    for key in sorted(params):
        ordered_params[key] = params[key]
    return ordered_params


def parse_response_line(line):
    k = line.find('=')
    key = line[:k]
    value = line[k + 1:]
    return key, value


def parse_response(text):
    params = {}
    for line in text.split('\n'):
        line = line.strip()
        if line != '':
            key, value = parse_response_line(line)
            params[key] = value
    return params


def response_signature(text, api_key):
    ordered_params = ordered_parameters(parse_response(text))
    # In opposition to request URls, the date here should not be escaped.
    text = '&'.join('%s=%s' % (k, v) for k, v in ordered_params.items())
    return hmac_sign_string(text, api_key)


def response_parameters_to_text(response_params):
    return ''.join(
        '%s=%s\r\n' % (k, v) for k, v in response_params.items()
    )


def http_text_response(response_params):
    return HttpResponse(response_parameters_to_text(response_params), content_type='text/plain')


def signed_http_text_response(response_params, api_key):
    text = response_parameters_to_text(response_params)
    signature = response_signature(text, api_key)
    text = 'h=%s\r\n%s' % (signature, text)
    return HttpResponse(text, content_type='text/plain')


def get_api_key_or_none(str_id):
    try:
        key_id = int(str_id)
    except ValueError:
        return

    try:
        return APIKey.objects.get(id=key_id)
    except APIKey.DoesNotExist:
        return


def hmac_sign_string(text, key):
    hashed = hmac.new(key, text.encode('utf-8'), hashlib.sha1)
    return base64.encodebytes(hashed.digest()).decode('utf-8')[:-1]


def ordered_parameters_string(query_dict, escape):
    sorted_keys = sorted(query_dict.keys())
    params = []
    for key in sorted_keys:
        for value in sorted(query_dict.getlist(key)):
            params.append((key, value))

    ordered_params = OrderedDict(params)
    if escape:
        return urlencode(ordered_params)
    else:
        return '&'.join('%s=%s' % (k, v) for k, v in ordered_params.items())


def hmac_verify_string(text, signature, key):
    expected_signature = hmac_sign_string(text, key)
    return hmac.compare_digest(signature, expected_signature)


def is_request_signature_valid(query_dict, api_key):
    # Reject if any key appears more than once
    if any(len(l) > 1 for _, l in query_dict.lists()):
        return False

    q = query_dict.copy()

    try:
        signatures = q.pop('h')
    except KeyError:
        return False
    signature = signatures[0]

    text = ordered_parameters_string(q, escape=True)
    return hmac_verify_string(text, signature, api_key)


class VerifyView(View):
    def get(self, request, *args, **kwargs):
        response = {
            't': datetime.datetime.utcnow().isoformat(),
        }

        required_fields = ['id', 'otp', 'nonce']
        if not all(name in request.GET for name in required_fields):
            response['status'] = ValidationStatus.MISSING_PARAMETER.value
            return http_text_response(response)

        token = request.GET.getlist('otp')[0]

        nonce = request.GET.getlist('nonce')[0]
        if '\r' in nonce or '\n' in nonce:
            response['status'] = ValidationStatus.MISSING_PARAMETER.value
            return http_text_response(response)

        response['nonce'] = nonce

        api_key = get_api_key_or_none(request.GET['id'])
        if api_key is None:
            response['status'] = ValidationStatus.NO_SUCH_CLIENT.value
            return http_text_response(response)

        key = base64.b64decode(api_key.key)

        if not is_request_signature_valid(request.GET, key):
            response['status'] = ValidationStatus.BAD_SIGNATURE.value
            return signed_http_text_response(response, key)

        if len(token) != 44 or '\r' in token or '\n' in token:
            response['status'] = ValidationStatus.BAD_OTP.value
            return signed_http_text_response(response, key)

        response['otp'] = token

        public_id = token[:12]
        devices = Device.objects.select_for_update().filter(public_id=public_id)
        with transaction.atomic():
            try:
                device = devices.get()
            except Device.DoesNotExist:
                response['status'] = ValidationStatus.BAD_OTP.value
                return signed_http_text_response(response, key)

            try:
                _, otp = decode_otp(token.encode('utf-8'), bytes.fromhex(device.key))
            except Exception:
                response['status'] = ValidationStatus.BAD_OTP.value
                return signed_http_text_response(response, key)

            response['sessionuse'] = otp.session
            response['sessioncounter'] = otp.counter
            response['timestamp'] = otp.timestamp

            if hexlify(otp.uid) != device.private_id.encode('utf-8'):
                response['status'] = ValidationStatus.BAD_OTP.value
                return signed_http_text_response(response, key)

            if otp.session < device.session_counter:
                response['status'] = ValidationStatus.REPLAYED_OTP.value
                return signed_http_text_response(response, key)

            if (otp.session == device.session_counter) and (otp.counter <= device.usage_counter):
                response['status'] = ValidationStatus.REPLAYED_OTP.value
                return signed_http_text_response(response, key)

            # OTP is valid; we update the counters:
            device.session_counter = otp.session
            device.usage_counter = otp.counter
            device.save()

        response['status'] = ValidationStatus.OK.value
        response['sl'] = 1
        return signed_http_text_response(response, key)



import base64
import secrets

from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models
from yubiotp.modhex import modhex

from yubival.validators import validate_modhex, validate_hex, LengthValidator


API_KEY_BYTE_LENGTH = 20
DEVICE_PUBLIC_ID_BYTE_LENGTH = 6
DEVICE_PRIVATE_ID_BYTE_LENGTH = 6
DEVICE_KEY_BYTE_LENGTH = 16


def generate_api_key():
    return base64.b64encode(secrets.token_bytes(API_KEY_BYTE_LENGTH)).decode('utf-8')


def generate_otp_key():
    return secrets.token_hex(16)


def generate_public_id():
    return modhex(secrets.token_bytes(DEVICE_PRIVATE_ID_BYTE_LENGTH)).decode('utf-8')


def generate_private_id():
    return secrets.token_hex(DEVICE_PRIVATE_ID_BYTE_LENGTH)


class APIKey(models.Model):
    label = models.CharField(
        max_length=64,
        unique=True,
    )
    key = models.CharField(
        max_length=4 * ((API_KEY_BYTE_LENGTH + 2) // 3),  # Max length after base64 conversion
        validators=[LengthValidator(4 * ((API_KEY_BYTE_LENGTH + 2) // 3))],
        default=generate_api_key,
    )
    date_created = models.DateTimeField(
        auto_now_add=True,
    )

    def __str__(self):
        return '%s (%d)' % (self.label, self.id)


class Device(models.Model):
    label = models.CharField(
        max_length=64,
        unique=True,
    )
    public_id = models.CharField(
        unique=True, max_length=2 * DEVICE_PUBLIC_ID_BYTE_LENGTH,
        validators=[LengthValidator(2 * DEVICE_PUBLIC_ID_BYTE_LENGTH), validate_modhex],
        default=generate_public_id,
    )
    private_id = models.CharField(
        unique=True, max_length=2 * DEVICE_PRIVATE_ID_BYTE_LENGTH,
        validators=[LengthValidator(2 * DEVICE_PRIVATE_ID_BYTE_LENGTH), validate_hex],
        default=generate_private_id,
    )
    key = models.CharField(
        max_length=2 * DEVICE_KEY_BYTE_LENGTH,
        validators=[LengthValidator(2 * DEVICE_KEY_BYTE_LENGTH), validate_hex],
        default=generate_otp_key,
    )
    session_counter = models.IntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(2**15 - 1)],
        default=0,
        editable=False,
    )
    usage_counter = models.IntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(255)],
        default=0,
        editable=False,
    )
    date_created = models.DateTimeField(
        auto_now_add=True,
    )

    def __str__(self):
        return '%s (%s)' % (self.label, self.public_id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yubival',
 'yubival.management',
 'yubival.management.commands',
 'yubival.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=2.2,<4.1', 'YubiOTP>=1.0.0,<2.0.0']

setup_kwargs = {
    'name': 'yubival',
    'version': '0.2.0rc1',
    'description': 'Django app that runs a standalone Yubikey validation server',
    'long_description': '![Default branch test status](https://github.com/bruot/yubival/actions/workflows/test.yml/badge.svg)\n![main branch coverage](https://codecov.io/gh/bruot/yubival/branch/main/graph/badge.svg?token=PNVDEEOHTU)\n\n\n# Yubival\n\nThis Django app runs a standalone Yubikey OTP validation server. It implements [version 2.0 of the validation protocol](https://developers.yubico.com/yubikey-val/Validation_Protocol_V2.0.html) limited to the case of a single validation server. YubiKey devices and server API keys can easily be managed in the Django admin site or via command line.\n\n\n## Installation\n\nYubival can be integrated to any existing Django project. Alternatively, you can create a new Django site to host your validation server. If unfamiliar with Django, please follow the instructions at "Create a new standalone validation server" below.\n\n\n### Add Yubival to an existing Django project\n\nInstall the package from PyPI:\n\n```\n$ pip install yubival\n```\n\nAdd `\'yubival\'` to the `INSTALLED_APPS` setting in settings.py. It is not required to enable the admin site. If you do, `INSTALLED_APPS` may look like:\n\n```\nINSTALLED_APPS = [\n    \'django.contrib.admin\',\n    \'django.contrib.auth\',\n    \'django.contrib.contenttypes\',\n    \'django.contrib.sessions\',\n    \'django.contrib.messages\',\n    \'django.contrib.staticfiles\',\n    \'yubival\',\n]\n```\n\nAdd the app URLs to the root urls.py file:\n\n```\nfrom django.contrib import admin\nfrom django.urls import path, include\n\nurlpatterns = [\n    # ...\n    path(\'admin/\', admin.site.urls),  # optional\n    path(\'\', include(\'yubival.urls\')),\n]\n```\n\nUpdate the database:\n\n```\npython manage.py migrate\n```\n\nWhen running the server, you should now be able to query the API at `/wsapi/2.0/verify`. When not providing any GET parameters, this returns a `MISSING_PARAMETER` status:\n\n```\nt=2021-10-29T08:31:11.885803\nstatus=MISSING_PARAMETER\n```\n\n\n### Create a new standalone validation server\n\nThis section explains how to setup a new Django site with Yubival. It was tested on a Debian 10 distribution, with Python 3.9 and Django 3.2.\n\nCreate a directory for the project:\n\n```\n$ mkdir myyubival\n$ cd myyubival\n```\n\nCreate a Python environment and activate it:\n\n```\n$ python3 -m venv venv\n$ source venv/bin/activate\n```\n\nInstall Django and Yubival:\n\n```\n$ pip install Django yubival\n```\n\nCreate a new Django project and browse to the newly created directory:\n\n```\n$ django-admin startproject myyubival\n$ cd myyubival\n```\n\nEdit the _./myyubival/settings.py_ file to add `\'yubival\'` to the `INSTALLED_APPS` setting:\n\n```\nINSTALLED_APPS = [\n    # ...\n    \'yubival\',\n]\n```\n\nMake the validation server URLs accessible by editing _./myyubival/urls.py_. Include the URLs from the Yubival app:\n\n```\nfrom django.contrib import admin\nfrom django.urls import path, include\nfrom django.views.generic.base import RedirectView\n\nurlpatterns = [\n    path(\'\', RedirectView.as_view(pattern_name=\'admin:index\')),\n    path(\'admin/\', admin.site.urls),\n    path(\'\', include(\'yubival.urls\')),\n]\n```\n\nFor convenience, we redirect above the website root to the admin area.\n\nBy default, Django will create a SQLite database located in a _db.sqlite3_ file in the project directory. To use other database engines, edit _./myyubival/settings.py_ to change the `DATABASES` setting; see the [Databases doc](https://docs.djangoproject.com/en/dev/ref/databases/). In both cases, run afterwards the following command to create the initial database tables:\n\n```\npython manage.py migrate\n```\n\nTo be able to use the admin site and manage Yubikey devices and server API keys, create an initial user account:\n\n```\n$ python manage.py createsuperuser\n```\n\nTo run the development web server, launch:\n\n```\n$ python manage.py runserver\n```\n\nThe website can now be accessed at http://127.0.0.1:8000/. It should show a "Page not found" error. The validation API is located at http://127.0.0.1:8000/wsapi/2.0/verify and the admin site interface at http://127.0.0.1:8000/admin/.\n\nWhile the `runserver` command above is an easy way to check your configuration and test Yubival, it should not be used to run the web server in production. Refer to the [deployment docs](https://docs.djangoproject.com/en/dev/howto/deployment/) to learn how to deploy your new myyubival site.\n\n\n## Commands usage\n\n### Getting help\n\nAll subcommands can print a detailed help with `--help`. Example:\n\n```\n$ python manage.py yubikey --help\n$ python manage.py yubikey add-existing --help\n```\n\n\n### API keys management\n\nServices that need to use the Yubival validation server require an API key which is a shared secret between Yubival and the service. The key is used to sign requests and responses to and from the validation server. API keys can be managed using the `manage.py apikey` command:\n\n```\n$ python manage.py apikey add service.example.com\nCreated: service.example.com (1):\n    id: 1\n    key: gnQ1sZWtRgCjm17waaiGHQptp8w=\n\n$ python manage.py apikey list\n1  service.example.com\n\n$ python manage.py apikey delete 1\nDeleted: service.example.com (1)\n```\n\n\n### Yubikey devices management\n\nYubiKeys can be added, listed and deleted using the commands below. To add a key, either use `manage.py yubikey add` that will automatically generate a public ID, a private ID and an AES key that you can use to configure a new Yubikey device, or use `manage.py yubikey add-existing` if you have a YubiKey for which you already know its parameters, including its secret key. In any case, make sure that the Yubival server will be the only validation server for the YubiKeys you register. If not, it would become possible to reuse OTP.\n\n```\n$ python manage.py yubikey add James\nCreated: James (cnfbfdinbblh):\n    Public ID: cnfbfdinbblh\n    Private ID: 1b935e02e095\n    AES key: fcea0ea12f97923ec4f952e0e170d419\n\n$ python manage.py yubikey add-existing Evelyn gkhcilelifuv 0123456789ab 00112233445566778899aabbccddeeff\nCreated: Evelyn (gkhcilelifuv)\n\n$ python manage.py yubikey list\ncnfbfdinbblh James\ngkhcilelifuv Evelyn\n\n$ python manage.py yubikey delete cnfbfdinbblh\nDeleted: James (cnfbfdinbblh)\n```\n',
    'author': 'Nicolas Bruot',
    'author_email': 'coremeltdown@bruot.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bruot/yubival',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from django.core.management.base import BaseCommand
from django.db import IntegrityError

from yubival.models import APIKey


class Command(BaseCommand):
    help = 'Manages API keys'
    requires_migrations_checks = True

    def _list(self):
        """Lists all API keys"""

        if APIKey.objects.count() == 0:
            return

        id_column_size = len(str(APIKey.objects.latest('id').id))
        row_format = '{:<%d} {:<}' % id_column_size

        for key in APIKey.objects.order_by('id'):
            self.stdout.write(row_format.format(key.id, key.label))

    def _add(self, label):
        """Adds an API key"""

        try:
            key = APIKey.objects.create(label=label)
        except IntegrityError as e:
            self.stdout.write(self.style.ERROR('Failed creating key: %s' % e.args[0]))
            return

        self.stdout.write(self.style.SUCCESS('Created: %s:' % str(key)))
        self.stdout.write('\tid: %s' % key.id)
        self.stdout.write('\tkey: %s' % key.key)

    def _delete(self, key_id):
        """Deletes an API key"""

        try:
            key = APIKey.objects.get(id=key_id)
        except APIKey.DoesNotExist:
            self.stdout.write(self.style.ERROR('API key id=%d does not exist.' % int(key_id)))
            return

        key_str = str(key)
        key.delete()
        self.stdout.write(self.style.SUCCESS('Deleted: %s' % key_str))

    def add_arguments(self, parser):
        subparsers = parser.add_subparsers(
            title='subcommands',
            dest='subcommand',
        )
        subparsers.required = True

        parser_add = subparsers.add_parser(
            'add',
            called_from_command_line=True,
            description='Creates an API key',
        )
        parser_add.add_argument('label', type=str, help='API key label')

        subparsers.add_parser(
            'list',
            called_from_command_line=True,
            description='Lists API keys',
        )

        parser_delete = subparsers.add_parser(
            'delete',
            called_from_command_line=True,
            description='Deletes an API key',
        )
        parser_delete.add_argument('id', type=int, help='API key ID')

    def handle(self, *args, **options):
        subcommand = options['subcommand']
        if subcommand == 'add':
            self._add(options['label'])
        elif subcommand == 'delete':
            self._delete(options['id'])
        else:  # subcommand == 'list'
            self._list()



import random

from django.core.management.base import BaseCommand
from django.db import IntegrityError
from yubiotp.modhex import modhex

from yubival.models import Device, DEVICE_PUBLIC_ID_BYTE_LENGTH, DEVICE_PRIVATE_ID_BYTE_LENGTH
from yubival.validators import argparse_type


class Command(BaseCommand):
    help = 'Manages YubiKey devices'
    requires_migrations_checks = True

    def _list(self):
        """Lists all registered YubiKeys"""

        row_format = '{:%d} {:<}' % DEVICE_PUBLIC_ID_BYTE_LENGTH

        for key in Device.objects.order_by('id'):
            self.stdout.write(row_format.format(key.public_id, key.label))

    def _add(self, label):
        """Registers a YubiKey by autogenerating device IDs and key"""
        try:
            device = Device.objects.create(label=label)
        except IntegrityError as e:
            self.stdout.write(self.style.ERROR('Failed creating device: %s' % e.args[0]))
            return

        self.stdout.write(self.style.SUCCESS('Created: %s:' % str(device)))
        self.stdout.write('\tPublic ID: %s' % device.public_id)
        self.stdout.write('\tPrivate ID: %s' % device.private_id)
        self.stdout.write('\tAES key: %s' % device.key)

    def _add_existing(self, label, public_id, private_id, key):
        """Registers an existing YubiKey"""
        try:
            device = Device.objects.create(
                label=label,
                public_id=public_id,
                private_id=private_id,
                key=key,
            )
        except IntegrityError as e:
            self.stdout.write(self.style.ERROR('Failed creating device: %s' % e.args[0]))
            return

        self.stdout.write(self.style.SUCCESS('Created: %s' % str(device)))

    def _delete(self, public_id):
        """Deletes a YubiKey"""
        try:
            device = Device.objects.get(public_id=public_id)
        except Device.DoesNotExist:
            self.stdout.write(self.style.ERROR('Device public_id=%s does not exist.' % public_id))
            return

        device_str = str(device)
        device.delete()
        self.stdout.write(self.style.SUCCESS('Deleted: %s' % device_str))

    def add_arguments(self, parser):
        subparsers = parser.add_subparsers(
            title='subcommands',
            dest='subcommand',
        )
        subparsers.required = True

        parser_add = subparsers.add_parser(
            'add',
            called_from_command_line=True,
            description='Registers a YubiKey with random IDs and secret key that can be uploaded to a YubiKey',
        )
        parser_add.add_argument('label', type=str, help='device label')

        parser_add_existing = subparsers.add_parser(
            'add-existing',
            called_from_command_line=True,
            description='Registers an already configured YubiKey',
        )

        random.seed(314159)
        example_public_id = modhex(bytearray([
            random.getrandbits(8) for _ in range(DEVICE_PUBLIC_ID_BYTE_LENGTH)
        ])).decode('utf-8')
        example_private_id = bytearray([
            random.getrandbits(8) for _ in range(DEVICE_PRIVATE_ID_BYTE_LENGTH)
        ]).hex()

        parser_add_existing.add_argument(
            'label',
            type=argparse_type(Device._meta.get_field('label')),
            help='device label',
        )
        parser_add_existing.add_argument(
            'public_id',
            type=argparse_type(Device._meta.get_field('public_id'), 'public_id_type'),
            help='public ID (%d-byte modhex such as "%s")' % (DEVICE_PUBLIC_ID_BYTE_LENGTH, example_public_id),
        )
        parser_add_existing.add_argument(
            'private_id',
            type=argparse_type(Device._meta.get_field('private_id'), 'private_id_type'),
            help='private ID (%d-byte hexadecimal such as "%s")' % (DEVICE_PRIVATE_ID_BYTE_LENGTH, example_private_id),
        )
        parser_add_existing.add_argument(
            'key',
            type=argparse_type(Device._meta.get_field('key'), 'key_type'),
            help='AES key (16-byte hexadecimal such as "00112233445566778899aabbccddeeff")',
        )

        subparsers.add_parser(
            'list',
            called_from_command_line=True,
            description='Lists YubiKeys',
        )

        parser_delete = subparsers.add_parser(
            'delete',
            called_from_command_line=True,
            description='Deletes a YubiKey',
        )
        parser_delete.add_argument(
            'public_id',
            type=str,
            help='YubiKey public ID (%d-byte modhex such as "%s")' % (DEVICE_PUBLIC_ID_BYTE_LENGTH, example_public_id),
        )

    def handle(self, *args, **options):
        subcommand = options['subcommand']
        if subcommand == 'add':
            self._add(options['label'])
        elif subcommand == 'add-existing':
            self._add_existing(options['label'], options['public_id'], options['private_id'], options['key'])
        elif subcommand == 'delete':
            self._delete(options['public_id'])
        else:  # subcommand == 'list'
            self._list()



from django.contrib import admin

from yubival.models import Device, APIKey


class APIKeyAdmin(admin.ModelAdmin):
    readonly_fields = (
        'date_created',
    )


class DeviceAdmin(admin.ModelAdmin):
    readonly_fields = (
        'session_counter',
        'usage_counter',
        'date_created',
    )


admin.site.register(APIKey, APIKeyAdmin)
admin.site.register(Device, DeviceAdmin)
